import os
from typing import Iterable

from mcap.reader import make_reader
from mcap_ros1.decoder import DecoderFactory as ROS1DecoderFactory
from mcap_ros2.decoder import DecoderFactory as ROS2DecoderFactory
from tqdm import tqdm

import native_rospy as rospy


# Custom progress bar class that inherits from tqdm.
class MCAPProgressBar(tqdm):
    def __init__(self, reader, disable, *args, **kwargs):
        """
        Initialize the progress bar using the reader's summary statistics.

        Parameters:
            reader: An MCAP reader instance.
            disable: A flag to disable the progress bar.
            *args, **kwargs: Additional arguments for tqdm.
        """
        self.reader = reader

        # If progress bar is enabled, set start and end times and initialize the tqdm progress bar.
        if not disable:
            statistics = self.reader.get_summary().statistics
            # Get the start and end times from the summary (in nanoseconds).
            self._start_time = statistics.message_start_time
            self._end_time = statistics.message_end_time
            self._now = self._start_time

            # Total progress is determined by the time range in seconds.
            total_seconds = (self._end_time - self._start_time) / 1e9
            super().__init__(total=total_seconds, unit="s", *args, **kwargs)
        else:
            # If disabled, instantiate tqdm with disable set to True.
            super().__init__(disable=True)

    def update_time(self, current_time):
        """
        Update the progress bar based on the current message time.

        Parameters:
            current_time: The current timestamp (in nanoseconds) from the message.
        """
        if not self.disable:
            # Calculate the difference in seconds and update the progress bar.
            delta_seconds = (current_time - self._now) / 1e9
            super().update(delta_seconds)
            self._now = current_time


def mcap_replay(mcap_file: os.PathLike, *, topics: Iterable[str] = None, show_pbar: bool = False):
    """
    Replay an MCAP file by reading messages and forwarding them to a simulated ROS middleware.

    Parameters:
        mcap_file: The path-like object representing the MCAP file.
        topics: An optional Iterable of topics to filter the messages; if None, process all topics.
        show_pbar: If True, display a progress bar for the replay process.

    The function reads each decoded message from the MCAP file, updates the simulated clock,
    publishes the message to the associated topic, and updates the progress bar.
    """
    # Get the default middleware instance from the native_rospy simulation.
    mw = rospy.get_default_middleware()

    # Open the MCAP file in binary mode.
    with open(mcap_file, "rb") as f:
        # Create an MCAP reader with the specified ROS1 decoder.
        reader = make_reader(f, decoder_factories=[ROS1DecoderFactory()])

        # Setup the progress bar with the reader and the show_pbar flag.
        with MCAPProgressBar(reader, disable=not show_pbar) as pbar:
            # Iterate through decoded messages; each yields a 4-tuple:
            # (schema, channel, message, ros_msg)
            for schema, channel, message, ros_msg in reader.iter_decoded_messages(topics):
                # Uncomment the below line for debugging output if needed.
                # print(f"{channel.topic} {schema.name} [{message.log_time}]: {ros_msg}")

                # Set the simulated clock to the current message's log time.
                mw.set_clock(message.log_time)
                # Publish the message to the appropriate topic.
                mw.publish(channel.topic, ros_msg)
                # Update the progress bar with the current message time.
                pbar.update_time(message.log_time)


# Make mcap_replay available for import from other modules.
__all__ = ["mcap_replay"]
