#!/usr/bin/env python
"""
A minimal imitation of rospy using native Python classes.
This module simulates basic ROS functionalities such as publishers, subscribers, timers,
and rates using a custom, simplified middleware.
"""

import enum
import threading
import time
from typing import Any, Callable

# --- Native Middleware Classes (provided and slightly enhanced) ---

# Constant representing an undefined time.
TIME_NOT_DEFINED = -1
# Conversion factor from seconds to nanoseconds.
S_TO_NS = 1_000_000_000


class TimeUnit(enum.Enum):
    """
    Enumeration for time units.
    """

    NSEC = "nsec"
    SEC = "sec"


class NativeMiddleware:
    """
    A simulation of the core middleware managing communication and timer callbacks.
    """

    def __init__(self):
        # Holds topic-specific callbacks for simulating publish/subscribe.
        self._callbacks: dict[str, list[Callable]] = {}
        # List of timer objects registered to this middleware.
        self._timers: list["Timer"] = []
        # Simulated clock value in nanoseconds.
        self._clock = TIME_NOT_DEFINED

    def publish(self, topic: str, msg: Any):
        """
        Simulate publishing a message to a topic by calling all registered callbacks.
        """
        if topic in self._callbacks:
            for callback in self._callbacks[topic]:
                callback(msg)

    def register_callback(self, topic: str, callback: Callable):
        """
        Register a callback function for a specific topic.
        """
        if topic not in self._callbacks:
            self._callbacks[topic] = []
        self._callbacks[topic].append(callback)

    def get_clock(self, unit: TimeUnit = TimeUnit.SEC):
        """
        Get the current simulated time in the requested unit (defaults to seconds).
        Returns None if the clock has not been set.
        """
        if self._clock == TIME_NOT_DEFINED:
            return None
        if unit == TimeUnit.SEC:
            return self._clock / S_TO_NS
        return self._clock

    def set_clock(self, time_value, unit: TimeUnit = TimeUnit.NSEC):
        """
        Set the simulated time. When the clock is updated, check all timers
        to see if they need to trigger.
        """
        # Convert to nanoseconds if necessary.
        time_value = int(time_value * 1e9) if unit == TimeUnit.SEC else int(time_value)
        self._clock = time_value
        # Check every timer to see if its period has passed.
        for timer in self._timers:
            if timer.is_period_passed(time_value):
                timer.callback()
                timer.last_called = time_value

    def register_timer(self, timer: "Timer"):
        """
        Add a timer to be managed by the middleware.
        """
        self._timers.append(timer)


# Create a singleton instance of the middleware to be shared across the module.
__middleware = None


def get_default_middleware():
    """
    Provides a default middleware instance (a simple singleton pattern).
    """
    global __middleware
    if __middleware is None:
        __middleware = NativeMiddleware()
    return __middleware


# --- Publisher, Subscriber, and other classes imitating the rospy API ---


class Publisher:
    """
    A simulated Publisher class that publishes messages to a topic.
    """

    __middleware = get_default_middleware()

    def __init__(self, topic, msg_type, *args, **kwargs):
        # Store the topic name; msg_type is simply accepted for compatibility.
        self.topic = topic

    def publish(self, msg):
        """
        Publish a message by delegating to the middleware's publish method.
        """
        self.__middleware.publish(self.topic, msg)


class Subscriber:
    """
    A simulated Subscriber class that registers a callback for a topic.
    """

    __middleware = get_default_middleware()

    def __init__(self, topic, msg_type, callback, *args, **kwargs):
        self.topic = topic
        # Register the callback with the middleware
        self.__middleware.register_callback(topic, callback)


class Timer:
    """
    A simulated Timer class that calls a callback at a fixed rate based on the middleware clock.
    """

    __middleware = get_default_middleware()

    def __init__(self, timer_period_sec: float, callback: Callable, *args, **kwargs):
        # Convert period to nanoseconds for internal consistency.
        self.period = int(timer_period_sec * S_TO_NS)
        self.callback = callback
        # Initially, the timer hasn't been called.
        self.last_called = TIME_NOT_DEFINED
        # Register this timer with the middleware.
        self.__middleware.register_timer(self)

    def is_period_passed(self, current_time):
        """
        Check if the timer period has passed since the last callback invocation.
        """
        return self.last_called == TIME_NOT_DEFINED or current_time - self.last_called >= self.period


class Rate:
    """
    Implements a simple rate mechanism.

    The sleep() method blocks until either the simulated time (provided by the middleware)
    has advanced for the desired period or, if no simulation time is defined,
    it falls back to real time.
    """

    __middleware = get_default_middleware()

    def __init__(self, rate: float):
        if rate <= 0:
            raise ValueError("Rate must be > 0")
        self.rate = rate
        # Calculate the sleep duration in seconds (1 / rate).
        self.sleep_duration = 1.0 / rate
        # A threading event used to block until the Timer wakes us up.
        self._event = threading.Event()
        # Register a Timer with the middleware to call the wake() method every period.
        self._timer = Timer(self.sleep_duration, self.wake)

    def wake(self):
        """
        The timer callback that releases the sleep() method by setting the event.
        """
        self._event.set()

    def sleep(self):
        """
        Attempt to sleep for the specified duration.
        If a simulated clock is available, wait until wake() gets called.
        Otherwise, use real time sleep.
        """
        if self.__middleware.get_clock(TimeUnit.SEC) is None:
            # No middleware clock defined; use standard time.sleep.
            time.sleep(self.sleep_duration)
        else:
            # Wait for the timer event, then clear it for the next period.
            self._event.wait()
            self._event.clear()


# --- Additional rospy-like functions and exceptions ---


def init_node(name, anonymous=False):
    """
    Initialize a node with the given name.
    In a real ROS system, this sets up the node; here it only prints a message.
    """
    print("Initialized node: {}".format(name))


def is_shutdown():
    """
    Check whether the system is shutting down.
    For our simulation, shutdown logic is not implemented.
    """
    return False


def loginfo(msg):
    """
    Print an informational log message.
    """
    print("[INFO] " + msg)


def spin():
    """
    Keep the node alive, simulating a spin loop.
    This loop simply sleeps until interrupted.
    """
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        # Gracefully exit on keyboard interruption.
        pass


class ROSException(Exception):
    """
    Base exception class for ROS clients.
    """

    pass


class ROSInterruptException(ROSException, KeyboardInterrupt):
    """
    Exception raised when an operation is interrupted (e.g., due to shutdown).
    Inherits from both ROSException and KeyboardInterrupt.
    """

    pass


# Expose key functions and classes to external modules.
__all__ = [
    "Publisher",
    "Subscriber",
    "Timer",
    "Rate",
    "init_node",
    "is_shutdown",
    "loginfo",
    "spin",
    "get_default_middleware",
    "ROSException",
    "ROSInterruptException",
]
